/*
 * This pthread module creates a thread that listens on a specified TCP port
 * for incoming connections.  Each connection received creates a new thread
 * to handle the session.  You may call ts_declare_port multiple times to
 * listen on more than one TCP port.  On thread exit, the connection is
 * closed.
 *
 * If a connection is received and no client contexts are available, the
 * listener thread will wait for up to three seconds for a context
 * to free up.
 *
 * int ts_declare_tcp_port ( port_num, client_limit, attr, thread, start );
 *	int start ( ctx, port, remote_address );
 * int ts_tcp_write ( ctx, buffer, length );
 * int ts_tcp_read ( ctx, buffer, maxlen, *length );
 * int ts_tcp_close ( ctx );
 * int ts_tcp_info ( local_port, remote_port, remote_address );
 * int ts_tcp_qio ( ctx, func, iosb, p1, p2, p3, p4, p5, p6 );
 *
 * Revised: 26-MAR-1994		Added CMU IP support.
 * Revised:  7-JUN-1994		Removed AST_SAFE stuff, DEC documented
 *				the cond_signal_int behaviour.
 * Revised: 20-JUL-1994		Cleanup use of startup_status.
 * Revised: 22-JUL-1994		Add code to stall for connection if no clients.
 */
#include "pthread_np.h"

/* Default to use UCX */
#ifndef CMUTCP
#ifndef TWGTCP
#ifndef MULTINET
#ifndef UCXTCP
#define UCXTCP
#endif
#endif
#endif
#endif

#include <stdlib.h>
#include <stdio.h>
#include <descrip.h>
#include <iodef.h>
int SYS$DASSGN();

#ifdef UCXTCP
#include <UCX$INETDEF.H>
#define sockaddr_in SOCKADDRIN
#endif
#ifdef MULTINET
#include "MULTINET_ROOT:[MULTINET.INCLUDE.SYS]TYPES.H"
#include "MULTINET_ROOT:[MULTINET.INCLUDE.NETINET]IN.H"
#include "MULTINET_ROOT:[MULTINET.INCLUDE.SYS]SOCKET.H"
#include "MULTINET_ROOT:[MULTINET.INCLUDE.VMS]INETIODEF.H"
#include "MULTINET_ROOT:[MULTINET.INCLUDE]errno.h"
#endif
#ifdef TWGTCP
#include "TWG$COMMON:[NETDIST.INCLUDE.SYS]TYPES.H"
#include "TWG$COMMON:[NETDIST.INCLUDE.SYS]SOCKET.H"
#include "TWG$COMMON:[NETDIST.INCLUDE.NETINET]IN.H"
#include "TWG$COMMON:[NETDIST.INCLUDE.VMS]INETIODEF.H"
#endif
#ifdef CMUTCP
#include <socket.h>
#include <in.h>
#define CMU_LISTEN_ADDRESS "0.0.0.0"
#define CMU_LISTEN_FLAGS 0
#define CMU_CLOSE_FLAGS 0
#define CMU_WRITE_FLAGS 0
#else
#define CMU_WRITE_FLAGS 0
#endif

#define ERROR_FAIL(s,text) if (s == -1) perror(text);
#define PORT_THREAD_STACK 4000

struct client_context {
    struct port_context *flink, *blink;
    int index, port_num;
    pthread_t thread;
    int status;
    pthread_cond_t io_done;
    int (*start)();		/* Application supplied startup routine */
    short chan;			/* Network channel */
    short clients_remaining;
    short iosb[4];

    struct sockaddr_in remote_address;     /* Remote address */
};
typedef struct client_context client_ctx_struct, *client_ctx;

struct port_context {
    struct port_context *flink, *blink;
    int port_num;
    int status;
    pthread_t thread;
#ifndef NOSTALL
    int connect_pending;		/* Connection is pending */
    pthread_cond_t client_available;	/* signals freelist is non-empty */
#endif
    pthread_attr_t *client_attr;	/* Attributes for created clients */
    short chan, fill;
    int (*start)();
    int client_count, client_limit;
    client_ctx_struct client_list;
    client_ctx free_clients;
};
typedef struct port_context port_ctx_struct, *port_ctx;
static port_ctx_struct port_list;
static int ts_startup_client ( client_ctx ctx );
port_ctx tserver_tcp_port_list;
/*
 * The following mutexes and conditions serialize access to resources:
 *    tcp_io		This mutex is used before SYS$QIO calls. 
 *			Thread-specific condition variables are then used
 *			in conjunction with an AST to wait on the I/O.
 *    tcp_ctl		This mutex is obtained when a thread needs to modify
 *			the client list or port list data structures.
 *    port_startup	Mutex to serialize access to port startup/init routine.
 *    port_startup_done	Condition used in conjuction with port_startup.
 */
static pthread_mutex_t tcp_io;		/* tcp/io data mutex */
static pthread_mutex_t tcp_ctl;		/* tcp context database mutex */
static pthread_mutex_t port_startup;	/* serialize port startup */
static pthread_cond_t port_startup_done;  /* condition */
static int startup_status;		/* Predicate */

static pthread_key_t port_key;		/* Private data for port listener */
static pthread_key_t client_key;	/* Private data for TCP client */
static void ts_init_mutexes();		/* forward reference */
static int ts_synch_ast(), ts_synch_io();

/*************************************************************************/
/* Create thread that listens on a particular TCP/IP port and starts
 * new thread for each accepted connection.  A caller supplied start routine
 * is called by the created thread to process the session.
 *
 * This routine can be called multiple times within an application to listen
 * on different ports.  Each port gets it's own client limit.
 *
 * User start (session) routine:
 *	int start ( void *ctx, int port_num, sockaddrin remote );
 */
int ts_declare_tcp_port ( 
    int port_num, 		/* Port number to listen on, host order */
    int client_limit, 		/* Limit on concurrent client threads allowed */
    pthread_attr_t *client_attr, /* Thread attributes for client threads */
    pthread_t *control_thread, 	/* Thread that listens for connects */
    int start() )		/* Start routine for new clients. */
{
    int ts_startup_port(), stacksize, status, pthread_create();
    port_ctx ctx;
    static pthread_once_t mutex_setup = pthread_once_init;
    pthread_attr_t port_attr;
    /*
     * Initialize globals for module, only once though.
     */
    pthread_once ( &mutex_setup, ts_init_mutexes );
    /*
     * Allocate port control block that becomes the thread-specific context
     * data for the new thread.
     */
    LOCK_C_RTL
    ctx = (port_ctx) malloc ( sizeof(port_ctx_struct) );
    UNLOCK_C_RTL
    /*
     * Link the control block into the global list, use mutex to synchronize
     * update of pointers.
     */
    pthread_mutex_lock ( &tcp_ctl );
    ctx->port_num = port_num;
    ctx->flink = &port_list;
    ctx->blink = port_list.blink;
    port_list.blink->flink = ctx;
    port_list.blink = ctx;
    pthread_mutex_unlock   ( &tcp_ctl );
    /*
     * Zero pertinent fields in control block.
     */
    ctx->status = 0;
    startup_status = 0;
    ctx->start = start;
    ctx->client_attr = client_attr;
    ctx->client_count = 0;
    ctx->client_limit = client_limit;
    /* 
     * Create thread and block until it finishes initializing so we can
     * check it's status.
     */
    status = pthread_attr_create ( &port_attr );
    ERROR_FAIL(status,"Port thread attribute create failure" )
    status = pthread_attr_setinheritsched ( &port_attr, PTHREAD_DEFAULT_SCHED );
    status = pthread_attr_setsched ( &port_attr, SCHED_FIFO );
    status = pthread_attr_setprio ( &port_attr, (PRI_FIFO_MIN+PRI_FIFO_MAX)/2 );
    status = pthread_attr_setstacksize ( &port_attr, PORT_THREAD_STACK );

    status = pthread_mutex_lock ( &port_startup );
    ERROR_FAIL(status,"Port startup lock failure" )

    status = pthread_create (control_thread, port_attr, ts_startup_port, ctx);
    ERROR_FAIL(status,"Port thread create failure" )

    while ( startup_status == 0 ) {
        status = pthread_cond_wait ( &port_startup_done, &port_startup );
        ERROR_FAIL(status,"Port startup wait failure" )
    }
    pthread_mutex_unlock ( &port_startup );

    return ctx->status;
}
/**************************************************************************/
/* Write data to thread's TCP/IP connection.
 */
int ts_tcp_write 
    ( client_ctx ctx,		/* Context passed to port start() function */
    char *buffer,		/* Data buffer to write */
    int length )		/* Number of chars to write */
{
    int SYS$QIO(), status, ts_synch_ast(), ts_synch_io(), remaining, seg;
    /*
     * Make $QIO call and block until AST signals us.
     */
    status = 1;
    for ( remaining = length; remaining > 0; remaining -= seg ) {
	/*
	 * Only write up to 1 K at a time.
	 */
	seg = remaining > 1024 ? 1024 : remaining;
        pthread_mutex_lock ( &tcp_io );
        status = SYS$QIO ( 0, ctx->chan, IO$_WRITEVBLK, &ctx->iosb,
	    ts_synch_ast, &ctx->io_done, buffer, seg, 0, CMU_WRITE_FLAGS, 
	    0, 0 );
	status = ts_synch_io ( status, &ctx->io_done, &ctx->iosb[0] );
        if ( (status&1) == 0 ) break;

#ifndef CMUTCP
	seg = ctx->iosb[1];	/* length actually sent */
#endif
	buffer = &buffer[seg];   /* Point to next unsent char */
    }
    return status;
}
/**************************************************************************/
/* Read data from thread's TCP/IP connection.
 */
int ts_tcp_read
    ( client_ctx ctx,		/* Context passed to port start() function */
    char *buffer,		/* Data buffer to write */
    int maxlen,			/* Size of buffer */
    int *length )		/* Number of chars to write */
{
    int SYS$QIO(), status, ts_synch_ast(), ts_synch_io();
    /*
     * Make $QIO call and block until AST signals us.
     */
    status = 1;
    pthread_mutex_lock ( &tcp_io );
    status = SYS$QIO ( 0, ctx->chan, IO$_READVBLK, &ctx->iosb,
	    ts_synch_ast, &ctx->io_done, buffer, maxlen, 0, 0, 0, 0 );
    status = ts_synch_io ( status, &ctx->io_done, &ctx->iosb[0] );
    if ( (status&1) == 1 ) {
	*length = ctx->iosb[1];
#ifdef TWGTCP
	/* WIN TCP return zero-length on EOF, convert it. */
	if ( *length == 0 ) status = 2160;
#endif
#ifdef MULTINET
	/* WIN TCP return zero-length on EOF, convert it. */
	if ( *length == 0 ) status = 2160;
#endif
    } else *length = 0;
    return status;
}
/**************************************************************************/
/* Perform synchronous disconnect of socket.  Do not deassign channel.
 * THe WIN/TCP and MULTINET interface does not have an expclicit close 
 * other than deassigning the channel, so do nothing for that case.
 */
int ts_tcp_close
    ( client_ctx ctx )		/* Context passed to port start() function */
{
    int SYS$QIO(), status, ts_synch_ast(), ts_synch_io();
    /*
     * Make $QIO call and block until AST signals us.
     */
    status = 1;
#ifdef UCXTCP
    pthread_mutex_lock ( &tcp_io );
    status = SYS$QIO ( 0, ctx->chan, IO$_DEACCESS, &ctx->iosb,
	    ts_synch_ast, &ctx->io_done, 0, 0, 0, 0, 0, 0 );
    status = ts_synch_io ( status, &ctx->io_done, &ctx->iosb[0] );
#endif
#ifdef CMUTCP
    /* Use _DELETE function to gracefully close connection, sending
     * all data. */
    pthread_mutex_lock ( &tcp_io );
    status = SYS$QIO ( 0, ctx->chan, IO$_DELETE, &ctx->iosb,
	    ts_synch_ast, &ctx->io_done, CMU_CLOSE_FLAGS, 0, 0, 0, 0, 0 );
    status = ts_synch_io ( status, &ctx->io_done, &ctx->iosb[0] );
#endif
    return status;
}
/**************************************************************************/
/*
 * Issue SYS$QIO on clients TCP channel and suspend until it completes.
 * Function value returned is either VMS status of $QIO or IOSB status.
 */
int ts_tcp_qio ( 
    client_ctx ctx, 		/* Context passed to port start() function */
    int func, 			/* $QIO function code/modifiers. */
    short iosb[4], 		/* copy of I/O status block returned by $QIO.*/
    int p1, int p2, int p3, 	/* Device dependant parameters to pass to */
    int p4, int p5, int p6 )	/* $QIO service. */
{
    int SYS$QIO(), status, ts_synch_ast(), ts_synch_io();
    /*
     * Make $QIO call and block until AST signals us.
     */
    pthread_mutex_lock ( &tcp_io );
    status = SYS$QIO ( 0, ctx->chan, func, &ctx->iosb,
	ts_synch_ast, &ctx->io_done, p1, p2, p3, p4, p5, p6 );
    status = ts_synch_io ( status, &ctx->io_done, &ctx->iosb[0] );
    if ( (status&1) == 0 ) return status;
    /*
     * Return copy of IOSB.
     */
    iosb[0] = ctx->iosb[0];
    iosb[1] = ctx->iosb[1];
    iosb[2] = ctx->iosb[2];
    iosb[3] = ctx->iosb[3];
    status = iosb[0];
    return status;
}
/**************************************************************************/
/* Do one-time initialization of static variables for this module, primarily
 * mutexes, condition variables and thread-specific context keys.
 */
static void port_rundown(), client_rundown();
static void ts_init_mutexes ( )
{
    int status;
    /*
     * Create mutexes and condition variables.
     */
    status = pthread_mutex_init ( &tcp_io, pthread_mutexattr_default );
    ERROR_FAIL(status,"Error creating tcp_io mutex" )

    status = pthread_mutex_init ( &tcp_ctl, pthread_mutexattr_default );
    ERROR_FAIL(status,"Error creating tcp_ctl mutex" )

    status = pthread_mutex_init ( &port_startup, pthread_mutexattr_default );
    ERROR_FAIL(status,"Error creating port_startup mutex" )

    status = pthread_cond_init ( &port_startup_done, pthread_condattr_default );
    ERROR_FAIL(status,"Error creating port_startup_done condition variable" )
    /* 
     * Create keys used to locate context blocks 
     */
    pthread_keycreate ( &port_key, port_rundown );
    ERROR_FAIL(status,"Error creating port listener context key" )

    pthread_keycreate ( &client_key, client_rundown );
    ERROR_FAIL(status,"Error creating tcp client context key" )
    /*
     * Initialize list head of port contexts created.  The pointers in
     * this structure are protected by the tcp_ctl mutex.
     */
    tserver_tcp_port_list = &port_list;		/* set global pointer */
    pthread_mutex_lock ( &tcp_ctl );
    port_list.flink = &port_list; port_list.blink = &port_list;
    port_list.port_num = 0;	/* flag list header */
    pthread_mutex_unlock ( &tcp_ctl );

}
/***************************************************************************/
/* Define VMS-specific routines to let us call system service routines
 * asynchronously.  Use ts_synch_ast as the ast argument to the system service
 * and a condition variable as the ast parameter.  The ts_synch_io routine
 * then blocks the thread waiting for the condition to be signalled.
 *
 * Note that ts_synch_io assumes tcp_io lock is held by the thread.  It is
 * released upon return.
 */
static int ts_synch_ast ( pthread_cond_t *done )
{
    int delta[2], SYS$SETIMR(), status;
    status = pthread_cond_signal_int_np ( done );

    return status;
}
static int ts_synch_io ( int status, pthread_cond_t *done, short *iosb )
{
    /*
     * Only do wait if system service status OK.
     */
    if ( (status&1) == 1 ) {
        /*
         * Loop until predicate (iosb nozero) is true.
         */
	do {
	    status = pthread_cond_wait ( done, &tcp_io );
	    if ( status == -1 ) break;
        } while ( *iosb == 0 );

	status = *iosb;
#ifdef CMUTCP
	if ( status == 44 ) {		/* abort status */
	    int *ptr;
	    ptr = (int *) iosb;
	    status = ptr[1];		/* IPACP status for abort */
	}
#endif
#ifdef SHOW_ERROR
	if ( status < 0 ) printf("synch_io has negative status: %d\n", status );
#endif
    }
    pthread_mutex_unlock ( &tcp_io );
    return status;
}
/***********************************************************************/
/*
 * Top level of thread for listening on ports
 */
int ts_startup_port ( port_ctx ctx )
{
    int status, stacksize, i, SYS$ASSIGN(), SYS$QIO();
    int client_chan, SYS$ASSIGN();
#ifdef UCXTCP
    $DESCRIPTOR(tcp_device,"UCX$DEVICE");
    struct sockname { long length; int *va; int *retadr; long list_end;
	struct SOCKADDRIN a; };
    struct sockname local, remote;
    struct sockchar { short protocol; char ptype, domain; } sockchar;
#endif
#ifdef MULTINET
    $DESCRIPTOR(tcp_device,"INET0:");
    struct sockaddr_in sock;
    struct { long length; struct sockaddr_in a; } remote;
#endif
#ifdef TWGTCP
    $DESCRIPTOR(tcp_device,"INET:");
    struct sockaddr_in sock;
    struct { long length; struct sockaddr_in a; } remote;
#endif
#ifdef CMUTCP
    $DESCRIPTOR(tcp_device,"IP:");
    struct cnx_info {
	unsigned short foreign_size, local_size;		/* name sizes */
	char foreign_host[128];
	unsigned long foreign_port;			/* Only low bytesused*/
	char local_host[128];				/* host names */
	unsigned long local_port;
	unsigned long local_addr, foreign_addr;		/* IP addresses */
    } connect_info;
    struct { long length; struct sockaddr_in a; } remote;
#endif
    struct iosb_t  { short status, count, d1, d2; } iosb;
    pthread_cond_t io_done;
    client_ctx new_client;
    static int marker = 0;
    /*
     * Create context key for this thread and init to our context block address.
     * The port rundown routine can then cleanup.
     */
    pthread_setspecific ( port_key, ctx );
    ctx->thread = pthread_self();
    ctx->client_list.flink = (port_ctx) &ctx->client_list;
    ctx->client_list.blink = (port_ctx) &ctx->client_list;
    ctx->client_list.index = 0;
#ifndef NOSTALL
    ctx->connect_pending = 0;
    status = pthread_cond_init 
		( &ctx->client_available, pthread_condattr_default );
    ERROR_FAIL(status,"Error creating port_startup_done condition variable" )
#endif
    /*
     * Pre-allocate client_limit number client contexts for the client threads.
     */
    ctx->free_clients = (client_ctx) 0;
    LOCK_C_RTL
    for ( i = 0; i < ctx->client_limit; i++ ) {
	client_ctx new;
	new = (client_ctx) malloc ( sizeof(client_ctx_struct) );
	new->index = marker+(ctx->client_limit-i);
	new->flink = (port_ctx) ctx->free_clients;
	ctx->free_clients = new;
    }
    marker += ctx->client_limit;	/* base for thread index numbers */
    UNLOCK_C_RTL
    /*
     * Assign control channel for listens.
     */
    ctx->chan = 0;
    ctx->status = SYS$ASSIGN ( &tcp_device, &ctx->chan, 0, 0 );
    if ( (ctx->status&1) == 1 ) {
	int opt_val;
        /*
	 * We found, device, create condition to synchronize I/O
	 */
	status = pthread_cond_init ( &io_done, pthread_condattr_default );
	/*
	 * Create socket and bind address that we will listen on.
	 */
#ifdef TWGTCP
        pthread_mutex_lock ( &tcp_io );
	ctx->status = SYS$QIO ( 0, ctx->chan, IO$_SOCKET, &iosb,
		ts_synch_ast, &io_done, 
		PF_INET, SOCK_STREAM, 0, 0, 0, 0 );
	ctx->status = ts_synch_io ( ctx->status, &io_done, &iosb.status );
	if ( (ctx->status&1) == 0 ) return ctx->status;
	/* Set socket option on socket to allow quick reuse of address */
	opt_val = 1;
        pthread_mutex_lock ( &tcp_io );
	ctx->status = SYS$QIO ( 0, ctx->chan, IO$_SETSOCKOPT, &iosb,
		ts_synch_ast, &io_done,
		SOL_SOCKET, SO_REUSEADDR, &opt_val, sizeof(opt_val), 0, 0 );
	ctx->status = ts_synch_io ( ctx->status, &io_done, &iosb.status );
	/* Set socket option to verify connection periodically */
	opt_val = 1;
        pthread_mutex_lock ( &tcp_io );
	ctx->status = SYS$QIO ( 0, ctx->chan, IO$_SETSOCKOPT, &iosb,
		ts_synch_ast, &io_done,
		SOL_SOCKET, SO_KEEPALIVE, &opt_val, sizeof(opt_val), 0, 0 );
	ctx->status = ts_synch_io ( ctx->status, &io_done, &iosb.status );
	if ( (ctx->status&1) == 0 ) return ctx->status;
	/* Bind socket to port number */
	sock.sin_family = PF_INET;
	sock.sin_port = ((ctx->port_num/256)&255) | ((ctx->port_num&255)<<8);
	sock.sin_addr.s_addr = 0;
	for(i=0;i<sizeof(sock.sin_zero);i++) sock.sin_zero[i] = 0;

        pthread_mutex_lock ( &tcp_io );
	ctx->status = SYS$QIO ( 0, ctx->chan, IO$_BIND, &iosb,
		ts_synch_ast, &io_done, 
		&sock, sizeof(sock), 0, 0, 0, 0 );
	ctx->status = ts_synch_io ( ctx->status, &io_done, &iosb.status );
	if ( (ctx->status&1) == 0 ) return ctx->status;
	/*
	 * Set backlog for connect requests.
	 */
        pthread_mutex_lock ( &tcp_io );
	ctx->status = SYS$QIO ( 0, ctx->chan, IO$_LISTEN, &iosb,
		ts_synch_ast, &io_done, 
		3, 0, 0, 0, 0, 0 );
	ctx->status = ts_synch_io ( ctx->status, &io_done, &iosb.status );
	if ( (ctx->status&1) == 0 ) return ctx->status;
#endif
#ifdef UCXTCP
	sockchar.protocol = 6; /* UCX$C_TCP; */
	sockchar.ptype =  1; /* UCX$C_STREAM; */
	sockchar.domain = 2; /* UCX$C_AF_INET; */

	local.length = sizeof(struct SOCKADDRIN);
	local.va = (int *) &local.a;
	local.retadr = 0;
	local.list_end = 0;
	local.a.SIN$W_FAMILY = UCX$C_AF_INET;
	local.a.SIN$W_PORT = ((ctx->port_num/256)&255) |
		((ctx->port_num&255)<<8);
	local.a.SIN$L_ADDR = 0;
	/*
	 * Create socket with parameters we want.
	 */
	{
	int flag;
	struct  itlst { short length, code;
	  union { struct itlst *lst; int *val; } addr; } olst[3];
	olst[0].code = UCX$C_SOCKOPT; olst[0].length = sizeof(struct itlst);
	olst[0].addr.lst = &olst[1];
	olst[1].length = 4; olst[1].code = UCX$C_REUSEADDR;
	olst[1].addr.val = &flag; flag = 1;
        pthread_mutex_lock ( &tcp_io );
	ctx->status = SYS$QIO ( 0, ctx->chan, IO$_SETMODE, &iosb,
		ts_synch_ast, &io_done, &sockchar, 0, 0, 0, 0, 0 );
	ctx->status = ts_synch_io ( ctx->status, &io_done, &iosb.status );
	tlog_putlog ( 5, "socket create status: !SL!/", ctx->status );
	if ( (ctx->status&1) == 0 ) return ctx->status;
        pthread_mutex_lock ( &tcp_io );
	ctx->status = SYS$QIO ( 0, ctx->chan, IO$_SETMODE, &iosb,
		ts_synch_ast, &io_done, 0, 0, 0, 0, &olst, 0 );
	ctx->status = ts_synch_io ( ctx->status, &io_done, &iosb.status );
	tlog_putlog ( 5, "socket setopt status: !SL!/", ctx->status );
	if ( (ctx->status&1) == 0 ) return ctx->status;
	/*
	 * Bind socket.
	 */
	else {
            pthread_mutex_lock ( &tcp_io );
	    ctx->status = SYS$QIO ( 0, ctx->chan, IO$_SETMODE, &iosb,
		ts_synch_ast, &io_done, 0, 0, &local, 3, 0, 0 );
	    ctx->status = ts_synch_io ( ctx->status, &io_done, &iosb.status );
	tlog_putlog ( 5, "socket bind status: !SL!/", ctx->status );
	    if ( (ctx->status&1) == 0 ) return ctx->status;
	}
	}
#endif    
#ifdef MULTINET
        pthread_mutex_lock ( &tcp_io );
	ctx->status = SYS$QIO ( 0, ctx->chan, IO$_SOCKET, &iosb,
		ts_synch_ast, &io_done, 
		AF_INET, SOCK_STREAM, 0, 0, 0, 0 );
	ctx->status = ts_synch_io ( ctx->status, &io_done, &iosb.status );
	if ( (ctx->status&1) == 0 ) return ctx->status;
	/* Set socket option on socket to allow quick reuse of address */
	opt_val = 1;
        pthread_mutex_lock ( &tcp_io );
	ctx->status = SYS$QIO ( 0, ctx->chan, IO$_SETSOCKOPT, &iosb,
		ts_synch_ast, &io_done,
		SOL_SOCKET, SO_REUSEADDR, &opt_val, sizeof(opt_val), 0, 0 );
	ctx->status = ts_synch_io ( ctx->status, &io_done, &iosb.status );
	/* Set socket option to verify connection periodically */
	opt_val = 1;
        pthread_mutex_lock ( &tcp_io );
	ctx->status = SYS$QIO ( 0, ctx->chan, IO$_SETSOCKOPT, &iosb,
		ts_synch_ast, &io_done,
		SOL_SOCKET, SO_KEEPALIVE, &opt_val, sizeof(opt_val), 0, 0 );
	ctx->status = ts_synch_io ( ctx->status, &io_done, &iosb.status );
	if ( (ctx->status&1) == 0 ) return ctx->status;
	/* Bind socket to port number */
	sock.sin_family = AF_INET;
	sock.sin_port = ((ctx->port_num/256)&255) | ((ctx->port_num&255)<<8);
	sock.sin_addr.s_addr = 0;
	for(i=0;i<sizeof(sock.sin_zero);i++) sock.sin_zero[i] = 0;

        pthread_mutex_lock ( &tcp_io );
	ctx->status = SYS$QIO ( 0, ctx->chan, IO$_BIND, &iosb,
		ts_synch_ast, &io_done, 
		&sock, sizeof(sock), 0, 0, 0, 0 );
	ctx->status = ts_synch_io ( ctx->status, &io_done, &iosb.status );
	if ( (ctx->status&1) == 0 ) return ctx->status;
	/*
	 * Set backlog for connect requests.
	 */
        pthread_mutex_lock ( &tcp_io );
	ctx->status = SYS$QIO ( 0, ctx->chan, IO$_LISTEN, &iosb,
		ts_synch_ast, &io_done, 
		3, 0, 0, 0, 0, 0 );
	ctx->status = ts_synch_io ( ctx->status, &io_done, &iosb.status );
	if ( (ctx->status&1) == 0 ) return ctx->status;
#endif
#ifdef CMUTCP
	SYS$DASSGN ( ctx->chan );
#endif
    }
    /* 
     * Acquire startup mutex so we know creating thread is waiting on us, then
     * set status variable and signal that setup is done 
     */
    status = pthread_mutex_lock ( &port_startup );
    if ( status == -1 ) perror ( "port_startup thread mutex" );
    startup_status = ctx->status;
    status = pthread_mutex_unlock ( &port_startup );
    status = pthread_cond_signal ( &port_startup_done );
    ctx->status = 0;		/* No longer starting */
    /*
     * Now  listen for incoming connections.
     */
    for ( ; ; ) {
	int length;
        client_chan = 0;
	status = SYS$ASSIGN ( &tcp_device, &client_chan, 0, 0 );
	if ( (status&1) == 0 ) return status;
#ifdef TWGTCP
	/*
	 * Listen for connect on newly assigned channel.
	 */
        pthread_mutex_lock ( &tcp_io );
	status = SYS$QIO ( 0, client_chan, IO$_ACCEPT, &iosb,
		ts_synch_ast, &io_done, 
		&remote, sizeof(remote), ctx->chan, 0, 0, 0 );
	status = ts_synch_io ( status, &io_done, &iosb.status );
	if ( (status&1) == 0 ) return status;
	
#endif
#ifdef UCXTCP
	remote.length = sizeof(struct SOCKADDRIN);
	remote.va = (int *) &remote.a;
	remote.retadr = &length;
	remote.list_end = 0;
	remote.a.SIN$W_FAMILY = UCX$C_AF_INET;
	remote.a.SIN$W_PORT = 0;	/* any port */
	remote.a.SIN$L_ADDR = 0;

        pthread_mutex_lock ( &tcp_io );
	status = SYS$QIO ( 0, ctx->chan, IO$_ACCESS|IO$M_ACCEPT, 
		&iosb, ts_synch_ast, &io_done, 
		0, 0, &remote, &client_chan, 0, 0 );
	status = ts_synch_io 
		( status, &io_done, &iosb.status );
	if ( (status&1) == 0 ) return status;
#endif
#ifdef MULTINET
	/*
	 * Listen for connect on newly assigned channel.
	 */
        pthread_mutex_lock ( &tcp_io );
	status = SYS$QIO ( 0, client_chan, IO$_ACCEPT, &iosb,
		ts_synch_ast, &io_done, 
		&remote, sizeof(remote), ctx->chan, 0, 0, 0 );
	status = ts_synch_io ( status, &io_done, &iosb.status );
	if ( (status&1) == 0 ) return status;
	
#endif
#ifdef CMUTCP
	/*
	 * Do passive connect on newly assigned channel.
	 */
        pthread_mutex_lock ( &tcp_io );
	status = SYS$QIO ( 0, client_chan, IO$_CREATE, &iosb,
		ts_synch_ast, &io_done,
		CMU_LISTEN_ADDRESS, 0, ctx->port_num, 0, 
		CMU_LISTEN_FLAGS, 0, 0);
	status = ts_synch_io ( status, &io_done, &iosb.status );
	if ( (status&1) == 0 ) return status;
	/*
	 * Get address info on this connection and build sockaddr structure.
	 */
        pthread_mutex_lock ( &tcp_io );
	status = SYS$QIO ( 0, client_chan, IO$_MODIFY, &iosb,
		ts_synch_ast, &io_done, 
		&connect_info, sizeof(connect_info), 0, 0, 0, 0 );
	status = ts_synch_io ( status, &io_done, &iosb.status );
	if ( (status&1) == 0 ) return status;
	remote.a.sin_family = PF_INET;
	remote.a.sin_port = ((connect_info.foreign_port/256)&255) | 
			((connect_info.foreign_port&255)<<8);
	remote.a.sin_addr.s_addr = connect_info.foreign_addr;
#endif
	/*
	 * Start new thread to handle connection.  See if client context
	 * available on free list.
	 */
	pthread_mutex_lock ( &tcp_ctl );
	new_client = ctx->free_clients;
#ifndef NOSTALL
	if ( ! new_client ) {
	    /*
	     * Give ourselves second chance to get client block, waiting
	     * up to 3 seconds. 
	     */
	    struct timespec delta, abstime;
	    delta.tv_sec = 3;
	    delta.tv_nsec = 0;
	    if ( 0 == pthread_get_expiration_np ( &delta, &abstime ) ) {
		tlog_putlog ( 0, "Thread list exhausted on port !SL, !%D!/", 
			ctx->port_num, 0 );
		ctx->connect_pending = 1;
	        pthread_cond_timedwait 
		    ( &ctx->client_available, &tcp_ctl, &abstime );
		ctx->connect_pending = 0;
		new_client = ctx->free_clients;
	    } else tlog_putlog ( 0, "Time computation problem!/" );
	}
#endif
	if ( new_client ) {
	    ctx->free_clients = (client_ctx) new_client->flink;
	    ctx->client_count++;
	    new_client->flink = (port_ctx) &ctx->client_list;
	    new_client->blink = ctx->client_list.blink;
	    ctx->client_list.blink = (port_ctx) new_client;
	    new_client->blink->flink = (port_ctx) new_client;
	    new_client->status = 0;
	    new_client->port_num = ctx->port_num;
	    new_client->clients_remaining = 
			ctx->client_limit - ctx->client_count;
	}
	pthread_mutex_unlock ( &tcp_ctl );

	if ( new_client ) {
	    int pthread_create();
	    new_client->chan = client_chan;
	    new_client->start = ctx->start;
	    new_client->remote_address = remote.a;

	    status = pthread_cond_init 
			( &new_client->io_done, pthread_condattr_default );

    	    status = pthread_create ( &new_client->thread, *(ctx->client_attr),
		ts_startup_client, new_client );
	    if ( status == -1 ) perror ( "error creating client thread" );
	} else {
	    LOCK_C_RTL
	    printf("Connect reject, no free client contexts\n");
	    UNLOCK_C_RTL
	    SYS$DASSGN ( client_chan );
	}
    }
    return status;
}

/***************************************************************************/
static int ts_startup_client ( client_ctx ctx )
{
    int status;

    pthread_setspecific ( client_key, ctx );

    status = (*ctx->start)(ctx, ctx->port_num, &ctx->remote_address, 
		ctx->index, ctx->clients_remaining );
    return status;
}
/***************************************************************************/
int ts_tcp_info ( int *local_port, int *remote_port, 
	unsigned int *remote_address )
{
    int status;
    client_ctx ctx;
    status = pthread_getspecific ( client_key, (pthread_addr_t *) &ctx );
    if ( status != 0 ) return status;
    if ( ctx == (client_ctx) 0 ) return -1;

    *local_port = ctx->port_num;
#ifdef UCXTCP
    *remote_port = ctx->remote_address.SIN$W_PORT;
    *remote_port = ((*remote_port>>8)&255) | ((*remote_port&255)<<8);
    *remote_address = ctx->remote_address.SIN$L_ADDR;
#else
#ifdef CMUTCP

    *remote_port = ctx->remote_address.sin_port;
    *remote_port = ((*remote_port>>8)&255) | ((*remote_port&255)<<8);
    *remote_address = ctx->remote_address.sin_addr.s_addr;
#else 	/* multinet and wollongong */
    *remote_port = ctx->remote_address.sin_port;
    *remote_port = ((*remote_port>>8)&255) | ((*remote_port&255)<<8);
    *remote_address = ctx->remote_address.sin_addr.s_addr;
#endif
#endif
    return 0;
}
/***************************************************************************/

static void client_rundown ( client_ctx ctx )
{
    int status, SYS$DASSGN();
    client_ctx t;
    port_ctx parent;
    if ( !ctx ) return;
    /* printf("running down tcp client, index=%d\n", ctx->index ); */
    ctx->status = 2;	/* flag as deleted */
    pthread_mutex_lock ( &tcp_ctl );
    for ( parent = port_list.flink; parent != &port_list; parent =
		parent->flink) {
	if ( parent->port_num == ctx->port_num ) {
	    /* Decrement counts */
	    --parent->client_count;
	    t = (client_ctx) ctx->flink;
	    t->blink = ctx->blink;
	    t = (client_ctx) ctx->blink; t->flink = ctx->flink;
	    break;
	}
    }
    status = SYS$DASSGN ( ctx->chan );
    pthread_cond_destroy ( &ctx->io_done );	/* cleanup condition */
    pthread_detach ( &ctx->thread );
    if ( parent == &port_list ) 
	printf("Client block not found, port\n", ctx->port_num );
    else {
	/* Place on free list, signal anybody waiting for client. */
	ctx->flink = (port_ctx) parent->free_clients;
	parent->free_clients = ctx;
#ifndef NOSTALL
	if ( !ctx->flink && parent->connect_pending ) {
	    tlog_putlog ( 0, "Signalling client available: !%T!/", 0 );
	    pthread_cond_signal ( &parent->client_available );
	}
#endif
    }
    pthread_mutex_unlock ( &tcp_ctl );
}

static void port_rundown (port_ctx ctx )
{
    port_ctx t;
    client_ctx client;
    int count;

    LOCK_C_RTL
    printf("running down tcp listen port, ctx address: %d\n", ctx );
    UNLOCK_C_RTL
    if ( !ctx ) return;
    if ( ctx->status ) {
	pthread_mutex_lock ( &port_startup );
	startup_status = ctx->status;
        ctx->status = 0;
	pthread_mutex_unlock ( &port_startup );
	pthread_cond_signal ( &port_startup_done );
    }

    /* Rundown clients */
    pthread_mutex_lock ( &tcp_ctl );
    for ( client = (client_ctx) ctx->client_list.flink;
	   client != &ctx->client_list; client = (client_ctx) client->flink ) {
	pthread_cancel ( client->thread );
    }
    pthread_mutex_unlock ( &tcp_ctl );
    count = ctx->client_count;

#ifndef NOSTALL
    pthread_cond_destroy ( &ctx->client_available );
#endif
    pthread_mutex_lock ( &tcp_ctl );
    t = ctx->flink; t->blink = ctx->blink;
    t = ctx->blink; t->flink = ctx->flink;
    pthread_mutex_unlock ( &tcp_ctl );

    SYS$DASSGN ( ctx->chan );

    pthread_detach ( &ctx->thread );

    LOCK_C_RTL
    for ( client = ctx->free_clients; client; client = ctx->free_clients ) {
	ctx->free_clients = (client_ctx) client->flink;
	free ( client );
    }
    free ( ctx );
    UNLOCK_C_RTL
}
